#!/usr/bin/env python2.7
"""    _  _  _  _  _        _                    _  _  _  _  _           _  _  _           _           _    
      (_)(_)(_)(_)(_)      (_)                  (_)(_)(_)(_)(_)       _ (_)(_)(_) _       (_)       _ (_)   
      (_)                  (_)                  (_)                  (_)         (_)      (_)    _ (_)      
      (_) _  _             (_)                  (_) _  _             (_)                  (_) _ (_)         
      (_)(_)(_)            (_)                  (_)(_)(_)            (_)                  (_)(_) _          
      (_)                  (_)                  (_)                  (_)          _       (_)   (_) _       
      (_)                  (_) _  _  _  _       (_) _  _  _  _       (_) _  _  _ (_)      (_)      (_) _    
      (_)                  (_)(_)(_)(_)(_)      (_)(_)(_)(_)(_)         (_)(_)(_)         (_)         (_)
   

                          Allineameto di immagini solari per correlazione incrociata.
                                      Giovanni Ceribella -- 11.05.2016

   Questo software esegue le seguenti operazioni:
     1) Converte le immagini da allineare in bianco e nero e ne effettua il binning 2x2.
     2) Inverte la scala di grigi, in modo che le  macchie solari o il  pianera in transito siano dei picchi
        luminosita'.
     3) Convolve le immagini con un  kernel gaussiano e stima il  valore massimo, cioe' il  punto piu' scuro
        delle immagini originali. I punti corrispondenti sono MAX1 e MAX2.
     4) Ritaglia un quadratino 128x128 attorno a questi due punti. Se fossero troppo vicini ai bordi sarebbe
        necessario modificare a mano il programma.
     5) Calcola la correlazione incrociata dei due quadratini, calcolandone prima le trasformate, lo spettro
        di potenza incrociato e antitrasformanto.
     6) Ottiene la posizione (xp,yp) del massimo della correlazione incrociata, pari alla  traslazione tra i
        quadratini approssimata al pixel.
     7) Interpola il profilo di  correlazione con  una gaussiana bidimensionale per ottenere una  precisione
        al di sotto del pixel (x0,y0).
     8) Trasla le immagini originale della quantita' -((MAX2-MAX1)+(x0,y0)) e le salva.

   Sintassi e argomenti:
     fleck referenza.img lista.txt [prefisso]

   Accetta qualunque formato sia interpretabile dalla Python Imaging Library.
"""


import sys

if(len(sys.argv)<4):
  sys.stderr.write("Sintassi:  "+sys.argv[0]+" referenza.img lista.txt prefisso\n")
  exit(1)

import time

import numpy as np

from scipy.ndimage import imread
from scipy.misc import imsave

from scipy.ndimage.filters import maximum_filter,gaussian_filter
from scipy.ndimage.interpolation import shift

import scipy.optimize as opt
import matplotlib.pyplot as plt
import gcim

fft = np.fft

plt.get_current_fig_manager().window.wm_geometry("900x300+100+100")
plt.gcf().set_tight_layout(True)

plt.ion()

import warnings
import matplotlib.cbook
warnings.filterwarnings("ignore",category=matplotlib.cbook.mplDeprecation)

def gaus2d((x, y), amplitude, xo, yo, sigma_x, sigma_y, theta):
    xo = float(xo)
    yo = float(yo)    
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo) + c*((y-yo)**2)))
    return g.ravel()


try:
  o1 = np.invert(imread(sys.argv[1],mode="L"))
  lista = np.loadtxt(sys.argv[2],dtype=str)
except:
  sys.stderr.write("Errore lettura 1.\n")
  exit(1)

########
im1 = (gcim.bin2x2Media(o1))[...,0].astype(np.uint16)*257

## Calcolo il massimo dopo una convoluzione gaussiana.
picco1 = maximum_filter(gaussian_filter(im1,sigma=2.0),size=2)
max1 = np.unravel_index(np.argmax(picco1),picco1.shape)
merc1 = o1[(2*max1[0]-64):(2*max1[0]+64),(2*max1[1]-64):(2*max1[1]+64)]

try:
  ## Calcolo le trasformate di fourier, lo spettro e la correlazione.
  fft1 = (fft.fft2(merc1,norm='ortho'))
except:
  sys.stderr.write("Errore: impossibile calcolare la trasformata della referenza.")
  exit(1)
########

num = lista.shape[0]
log = open("flecken.log","a")

log.write("######### "+str(time.ctime())+"\n")
log.write("#### REFERENZA = "+sys.argv[1]+"\n")
log.write("#### LISTA     = "+sys.argv[2]+"\n")
log.write("#### PREFISSO  = "+sys.argv[3]+"\n")
log.write("## N. IMMAGINI = "+str(num)+"\n\n")

q3 = plt.subplot(1,3,3)
q2 = plt.subplot(1,3,2)
q1 = plt.subplot(1,3,1)
plt.imshow(merc1,extent=(2*max1[1]-64,2*max1[1]+64,2*max1[0]-64,2*max1[0]+64))
q1.set_title("REF. "+sys.argv[1]) 


for j in range(num):
 
  nomefilej = lista[j]
  log.write(str(j)+"  ALIGN: reference "+sys.argv[1]+" ; image "+nomefilej+'\n')
  
  try:
    o2 = np.invert(imread(nomefilej,mode="L"))
  except:
    sys.stderr.write(nomefilej+": errore lettura. Salto alla successiva.\n")
    log.write(nomefilej+": errore lettura. Salto alla successiva.\n")
    continue

  im2 = (gcim.bin2x2Media(o2))[...,0].astype(np.uint16)*257
  
  ## Calcolo il massimo dopo una convoluzione gaussiana.
  picco2 = maximum_filter(gaussian_filter(im2,sigma=2.0),size=2)
  
  max2 = np.unravel_index(np.argmax(picco2),picco2.shape)
  log.write(" MAX1 = "+repr(max1)+'\n')
  log.write(" MAX2 = "+repr(max2)+'\n\n')

  
  ## Definisco due quadratini attorno ai massimi
  merc2 = o2[(2*max2[0]-64):(2*max2[0]+64),(2*max2[1]-64):(2*max2[1]+64)]
  
  ## Utile per il debug: (axis=0 -> y ; axis=1 -> x)
  ## merc2 = np.roll(merc1,20,axis=0)
  
  q2 = plt.subplot(1,3,2)
  plt.imshow(merc2,extent=(2*max2[1]-64,2*max2[1]+64,2*max2[0]-64,2*max2[0]+64))
  q2.set_title(nomefilej)
  
  try:
    ## Calcolo le trasformate di fourier, lo spettro e la correlazione.
    fft2 = (fft.fft2(merc2,norm='ortho'))
  
  
    power = (np.conj(fft1)*(fft2))
    corr = np.real(fft.ifft2(power,norm='ortho'))
  
    corr = fft.fftshift(corr)
  
    corr -= corr.min()
    corr /= corr.max()
    bcor = True
  except:
    sys.stderr.write(" !! "+nomefilej+": CALCOLO CROSS-CORRELAZIONE FALLITO. Uso i massimi.\n")
    log.write(" !! "+nomefilej+": CALCOLO CROSS-CORRELAZIONE FALLITO. Uso i massimi.\n")
    bcor = False
  
  if(bcor):
    yp,xp = np.unravel_index(np.argmax(corr),corr.shape)
  
    q3 = plt.subplot(1,3,3)
    q3.set_title("Correlation")
    plt.imshow(corr,extent=(-64,63,-64,63))
  
    plt.plot(xp-64,64-yp,"bo")
  
    log.write(" xp = "+str(xp-64)+'\n')
    log.write(" yp = "+str(yp-64)+'\n\n')

  
    piniz = (1,xp,yp,10,10,0)
  
    x = np.linspace(0,127,128)
    y = np.linspace(0,127,128)
  
    x, y = np.meshgrid(x, y)
  
    try:
      ## Fit Gaussiano per precisione sub-pixel.
      popt, pcov = opt.curve_fit(gaus2d, (x, y), corr.ravel(), p0=piniz)
      bfit = True
    except:
      sys.stderr.write(" !! "+nomefilej+": FIT GAUSSIANO FALLITO. Uso le stime intere.\n")
      log.write(" !! "+nomefilej+": FIT GAUSSIANO FALLITO. Uso le stime intere.\n")
      bfit = False
  
    if(bfit):
      perr = np.sqrt(np.diag(pcov))
  
      log.write( " A  = %0.3f +/- %0.3f\n" % (popt[0],perr[0])   )
      log.write( " x0 = %0.3f +/- %0.3f\n" % (popt[1]-64,perr[1]))
      log.write( " y0 = %0.3f +/- %0.3f\n" % (popt[2]-64,perr[2]))
      log.write( " sx = %0.3f +/- %0.3f\n" % (popt[3],perr[3])   )
      log.write( " sy = %0.3f +/- %0.3f\n" % (popt[4],perr[4])   )
      log.write( " th = %0.3f +/- %0.3f\n\n" % (popt[5],perr[5])   )

  
      fitg = gaus2d((x,y), *popt).reshape((128,128))
  
      plt.contour(x-64, 64-y, fitg, 8, colors='w')
      plt.plot(popt[1]-64,64-popt[2],"ro")
  
      dx = (popt[1]-64)  +  2*(max2[1]-max1[1])
      dy = (popt[2]-64)  +  2*(max2[0]-max1[0])
    else:
      ## Se il fit va male, uso la precisione al pixel.
      dx = xp + 2*(max2[1]-max1[1])
      dy = yp + 2*(max2[0]-max1[0])
  else:
    ## Se la correlazione va male, uso solo i massimi.
    dx = 2*(max2[1]-max1[1])
    dy = 2*(max2[0]-max1[0])
  
  log.write(" Allineamento trovato:\n")
  log.write("  DX = "+str(dx)+'\n')
  log.write("  DY = "+str(dy)+'\n')
  
  nomeoutj = sys.argv[3]+nomefilej

  log.write(" Salvo "+nomeoutj+" ...")
  
  o2shift = shift(imread(nomefilej,mode="RGB"),(-dy,-dx,0),order=3)
  
  try:
    imsave(nomeoutj,o2shift)
    log.write(" fatto.\n")
  except:
    log.write("!!\n")
    log.write("Errore scrittura.\n")
    sys.stderr.write("Errore scrittura.\n")
    exit(2)

  plt.pause(0.0001)
  q2.clear()
  q3.clear()


imsave(sys.argv[3]+sys.argv[1],imread(sys.argv[1],mode="RGB"))

exit(0)
